"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MorseString = void 0;
const Helpers_1 = require("../Common/Helpers");
const MorseCharacter_1 = require("./MorseCharacter");
// MorseString represents a string of multiple morse characters.  It allows for a longer representation
// to be converted to a single string and allows for sentence-level transforms such as reversing
// the order of tokens.
class MorseString {
    static CHARACTER_DIVIDER = '/';
    static WORD_DIVIDER = ' ';
    constructor(morse = '', characterDivider = MorseString.CHARACTER_DIVIDER, wordDivider = MorseString.WORD_DIVIDER) {
        // The dividers should be single characters which don't clash with the other string content
        Helpers_1.Helpers.assertMsg(characterDivider.length === 1, 'Divider must be a single character');
        Helpers_1.Helpers.assertMsg(wordDivider.length === 1, 'Divider must be a single character');
        Helpers_1.Helpers.assertMsg(characterDivider !== wordDivider, 'Dividers must be different from each other');
        Helpers_1.Helpers.assertMsg(characterDivider !== MorseCharacter_1.MorseCharacter.DASH, 'Character divider must not be a reserved value');
        Helpers_1.Helpers.assertMsg(characterDivider !== MorseCharacter_1.MorseCharacter.DOT, 'Character divider must not be a reserved value');
        Helpers_1.Helpers.assertMsg(characterDivider !== MorseCharacter_1.MorseCharacter.RESERVED_DIVIDER, 'Character divider must not be a reserved value');
        Helpers_1.Helpers.assertMsg(wordDivider !== MorseCharacter_1.MorseCharacter.DASH, 'Word divider must not be a reserved value');
        Helpers_1.Helpers.assertMsg(wordDivider !== MorseCharacter_1.MorseCharacter.DOT, 'Word divider must not be a reserved value');
        Helpers_1.Helpers.assertMsg(wordDivider !== MorseCharacter_1.MorseCharacter.RESERVED_DIVIDER, 'Word divider must not be a reserved value');
        const words = morse.split(wordDivider).filter(w => w.length > 0);
        // Discard any empty characters (caused by trailing separator)
        const wordCharacters = words.map(w => w.split(characterDivider).filter(wc => wc.length > 0));
        this._words = wordCharacters.map(subarray => subarray.map(wc => new MorseCharacter_1.MorseCharacter(wc)));
    }
    // reverse reverses the order of all tokens in the string (including separators), such as would be
    // the case if you were reading the data for a puzzle backwards.
    reverse() {
        for (let i = 0; i < this._words.length; i++) {
            for (const c of this._words[i]) {
                // Reverse the dots/dashes in each letter
                c.reverse();
            }
            // Reverse the letters in each word
            this._words[i] = this._words[i].reverse();
        }
        // Reverse the order of the words
        this._words = this._words.reverse();
        return this;
    }
    // invertDotsAndDashes switches all dots and dashes in the input, such as would be the case if
    // you had two ambiguous symbols for dot/dash and selected the wrong mapping.
    invertDotsAndDashes() {
        for (const word of this._words) {
            for (const c of word) {
                c.invertDotsAndDashes();
            }
        }
        return this;
    }
    toString() {
        let s = '';
        for (let i = 0; i < this._words.length; i++) {
            if (i > 0) {
                s += ' '; // This is a new word
            }
            for (const c of this._words[i]) {
                s += c.toString() || '?';
            }
        }
        return s;
    }
    _words;
}
exports.MorseString = MorseString;
//# sourceMappingURL=MorseString.js.map