"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.WordSearchSolver = exports.WordSearchSpaceTreatment = void 0;
const trie = require("trie-prefix-tree");
const Result_1 = require("./Result");
const WordSearchDirection_1 = require("./WordSearchDirection");
var WordSearchSpaceTreatment;
(function (WordSearchSpaceTreatment) {
    WordSearchSpaceTreatment[WordSearchSpaceTreatment["None"] = 0] = "None";
    WordSearchSpaceTreatment[WordSearchSpaceTreatment["RemoveWithinPuzzle"] = 1] = "RemoveWithinPuzzle";
    WordSearchSpaceTreatment[WordSearchSpaceTreatment["RemoveAll"] = 2] = "RemoveAll";
})(WordSearchSpaceTreatment || (exports.WordSearchSpaceTreatment = WordSearchSpaceTreatment = {}));
class WordSearchSolver {
    // The grid of letters which makes up the wordsearch puzzle
    _matrix;
    // Transpositions to apply when finding words
    _directions;
    // Potential words to find in the letter grid
    _targets;
    // Should we search for words not following a line
    _canBend;
    // What should we do with spaces
    _spaceTreatment;
    constructor() {
        // Use empty grid and list by default
        this._matrix = [[]];
        this._targets = trie([]);
        // Use both sets of directions by default
        this._directions = [
            // Cardinal directions
            [0, 1],
            [-1, 0],
            [1, 0],
            [0, -1],
            // Diagonal
            [1, 1],
            [-1, -1],
            [1, -1],
            [-1, 1],
        ];
        this._canBend = false;
        this._spaceTreatment = WordSearchSpaceTreatment.None;
    }
    setWords(words) {
        this._targets = trie([]);
        for (const full of words) {
            if (full === null || typeof full === 'undefined') {
                throw new Error('Invalid input in WordSearchSolver.findWords()');
            }
            const word = full.trim();
            if (word !== '') {
                this._targets.addWord(word);
            }
            else {
                throw new Error('Cannot find an empty string in the wordsearch');
            }
        }
    }
    parseGrid(gridInputText) {
        const lines = gridInputText.split(/\r?\n/);
        const grid = [];
        for (const line of lines) {
            grid.push(line.split(''));
        }
        this.setGrid(grid);
    }
    setGrid(matrix) {
        this._matrix = matrix;
        this.checkSpaces();
    }
    setDirections(direction) {
        this._directions = [];
        if (direction === WordSearchDirection_1.WordSearchDirection.Cardinal ||
            direction === WordSearchDirection_1.WordSearchDirection.CardinalAndDiagonal) {
            this._directions = this._directions.concat([
                [0, 1],
                [-1, 0],
                [1, 0],
                [0, -1],
            ]);
        }
        if (direction === WordSearchDirection_1.WordSearchDirection.Diagonal ||
            direction === WordSearchDirection_1.WordSearchDirection.CardinalAndDiagonal) {
            this._directions = this._directions.concat([
                [1, 1],
                [-1, -1],
                [1, -1],
                [-1, 1],
            ]);
        }
    }
    setSpaceTreatment(treatment) {
        this._spaceTreatment = treatment;
        if (this._matrix.length > 0) {
            this.checkSpaces();
        }
    }
    setCanBend(canBend) {
        this._canBend = canBend;
    }
    findWords() {
        const results = [];
        const numRows = this._matrix.length;
        for (let yIdx = 0; yIdx < numRows; yIdx++) {
            const lineLength = this._matrix[yIdx].length;
            for (let xIdx = 0; xIdx < lineLength; xIdx++) {
                const p = { x: xIdx, y: yIdx };
                const pointResults = this.search(p);
                results.push(...pointResults);
            }
        }
        return results;
    }
    search(start) {
        const results = [];
        if (this._canBend) {
            const history = [];
            const dfsResults = this.dfsCheck(start, new Set(), history);
            results.push(...dfsResults);
        }
        else {
            for (const translation of this._directions) {
                const directionalResults = this.lineCheck(start, translation);
                results.push(...directionalResults);
            }
        }
        return results;
    }
    dfsCheck(start, visited, history) {
        const results = [];
        if (visited.has(start)) {
            return results;
        }
        visited.add(start);
        history.push(start);
        let currentString = '';
        for (let i = 0; i < history.length; i++) {
            const pt = history[i];
            currentString = currentString + this._matrix[pt.y][pt.x];
        }
        const wordsWithPrefix = this._targets.getPrefix(currentString);
        if (wordsWithPrefix.length === 0) {
            visited.delete(start);
            history.pop();
            return results;
        }
        if (wordsWithPrefix.indexOf(currentString) !== -1) {
            const foundWord = new Result_1.Result(currentString, history);
            results.push(foundWord);
        }
        for (const translation of this._directions) {
            const next = {
                x: start.x + translation[0],
                y: start.y + translation[1],
            };
            if (this.isInBounds(next)) {
                results.push(...this.dfsCheck(next, visited, history));
            }
        }
        visited.delete(start);
        history.pop();
        return results;
    }
    lineCheck(start, direction) {
        const results = [];
        // Working set
        let currentPoint = start;
        let currentString = '';
        const pointHistory = [];
        while (this.isInBounds(currentPoint)) {
            currentString =
                currentString + this._matrix[currentPoint.y][currentPoint.x];
            // Get the candidates with the current prefix string
            const wordsWithPrefix = this._targets.getPrefix(currentString);
            // No point in going on, we've run out of possibilities
            if (wordsWithPrefix.length === 0) {
                break;
            }
            const p = { x: currentPoint.x, y: currentPoint.y };
            pointHistory.push(p);
            // Is a candidate an exact match for the current search string? Save it.
            if (wordsWithPrefix.indexOf(currentString) !== -1) {
                const foundWord = new Result_1.Result(currentString, pointHistory);
                results.push(foundWord);
            }
            const next = {
                x: currentPoint.x + direction[0],
                y: currentPoint.y + direction[1],
            };
            currentPoint = next;
        }
        return results;
    }
    isInBounds(point) {
        if (point.y < 0 || point.y >= this._matrix.length) {
            return false;
        }
        if (point.x < 0 || point.x >= this._matrix[point.y].length) {
            return false;
        }
        return true;
    }
    checkSpaces() {
        if (this._matrix.length === 0 ||
            this._spaceTreatment === WordSearchSpaceTreatment.None) {
            return;
        }
        const nextMatrix = [];
        for (let yIdx = 0; yIdx < this._matrix.length; yIdx++) {
            const line = [];
            let inPuzzle = false;
            for (let xIdx = 0; xIdx < this._matrix[yIdx].length; xIdx++) {
                const letter = this._matrix[yIdx][xIdx];
                if (letter !== ' ') {
                    inPuzzle = true;
                }
                if (this._spaceTreatment === WordSearchSpaceTreatment.RemoveAll &&
                    letter !== ' ') {
                    line.push(this._matrix[yIdx][xIdx]);
                }
                else if (letter !== ' ' && inPuzzle) {
                    line.push(this._matrix[yIdx][xIdx]);
                }
            }
            if (line.length > 0) {
                nextMatrix.push(line);
            }
        }
        this._matrix = nextMatrix;
    }
}
exports.WordSearchSolver = WordSearchSolver;
//# sourceMappingURL=WordSearchSolver.js.map